import pytest
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split

from swi_ml import (
    logger,
    set_automatic_fallback,
    set_backend,
    set_logging_level,
)
from swi_ml.classification import (
    LogisticRegressionGD,
)

set_logging_level("DEBUG")
logger.setLevel("DEBUG")

X, Y = make_classification(n_samples=1000, n_features=5, random_state=42)

X_train, X_test, Y_train, Y_test = train_test_split(
    X, Y, test_size=1 / 3, random_state=0
)


def test_logistic_regression():
    num_iters = 10
    lr = 0.05

    # set numpy backend first
    set_backend("numpy")
    model_np = LogisticRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
    )

    with pytest.raises(ValueError):
        model_np.fit([], [])

    model_np.fit(X_train, Y_train)
    Y_pred_np = model_np.predict(X_test)

    with pytest.raises(NotImplementedError):
        wrong_initialiser_model = LogisticRegressionGD(
            num_iterations=num_iters,
            learning_rate=lr,
            initialiser="wrong_initialiser",
        )
        wrong_initialiser_model.fit(X_train, Y_train)

    # set cupy backend
    set_backend("cupy")
    model_cp = LogisticRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
    )

    model_cp.fit(X_train, Y_train)

    Y_pred_cp = model_cp.predict(X_test)

    assert len(model_np.history) == len(model_cp.history) == num_iters
    assert Y_pred_np.shape == Y_pred_cp.shape

    # switch backend back to numpy
    set_backend("numpy")

    model = LogisticRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
        multiply_factor=100,
        l1_ratio=0.5,
        normalize=True,
    )

    model.fit(X_train, Y_train)
    Y_pred = model.predict(X_test, probability=True)

    assert Y_pred is not None



from swi_ml import activations
from swi_ml.regression.linear_regression import (
    _BaseRegression,
    L1_L2Regularisation,
)


class LogisticRegressionGD(_BaseRegression):
    def __init__(
        self,
        num_iterations: int,
        learning_rate: float,
        multiply_factor=None,
        l1_ratio=None,
        normalize=False,
        initialiser="uniform",
        verbose=None,
    ):
        self.activation = activations.Sigmoid()
        regularisation = L1_L2Regularisation(
            multiply_factor=multiply_factor, l1_ratio=l1_ratio
        )
        super().__init__(
            num_iterations,
            learning_rate,
            normalize,
            regularisation,
            initialiser,
            verbose,
        )

    def predict(self, X, probability=False):
        """
        Given an input array X, it returns the prediction array
        (GPU array if CuPy backend is enabled) after inferencing
        """
        if probability:
            return self._predict(self._predict_preprocess(X))
        else:
            activated_pred = self.activation.activate(
                self._predict(self._predict_preprocess(X))
            )
            return self.backend.where(activated_pred > 0.5, 1, 0)

    def _predict(self, X):
        return self.activation.activate(
            self.backend.asarray(X).dot(self.W) + self.b
        )



# changenote: If you're viewing this commit, then it means the name of the
# library is changed from cu-ml to swi-ml to avoid confusion with cuML, a
# library from RapidsAI (https://github.com/rapidsai/cuml)

import logging
import sys

from .logs import INFOFORMATTER, DEBUGFORMATTER

_fallback_to_numpy = False


def set_automatic_fallback(boolean):
    global _fallback_to_numpy
    _fallback_to_numpy = boolean


logger = logging.getLogger(__name__)

# create the stream handler
_ch = logging.StreamHandler()


def set_logging_level(level):
    # sets the handler info
    _ch.setLevel(level)
    # set the handler formatting
    if level == "DEBUG":
        _ch.setFormatter(logging.Formatter(DEBUGFORMATTER))
    else:
        _ch.setFormatter(logging.Formatter(INFOFORMATTER))


set_logging_level("INFO")
# adds the handler to the global variable: log
logger.addHandler(_ch)
logger.setLevel(logging.WARN)


from .backend import _Backend

# backend throughout the library, to be set by user
_global_backend = _Backend()


def set_backend(module_backend):
    _global_backend.set_backend(module_backend)


from .utils import (
    activations,
    distributions,
    manipulations,
    regularisers,
)

for module in (activations, distributions, manipulations, regularisers):
    full_name = "{}.{}".format(__package__, module.__name__.rsplit(".")[-1])
    sys.modules[full_name] = sys.modules[module.__name__]



from itertools import combinations_with_replacement

try:
    import numpy as np
except ImportError:
    # exception already handled in backend
    pass


def index_combinations(degree, num_features):
    combinations = [
        combinations_with_replacement(range(num_features), i)
        for i in range(0, degree + 1)
    ]
    flat_combinations = [item for sublist in combinations for item in sublist]
    return flat_combinations


def transform_polynomial(X, degree):
    n_samples, n_features = np.shape(X)
    combinations = index_combinations(degree, n_features)
    n_output_features = len(combinations)

    X_new = np.empty((n_samples, n_output_features))

    for i, index_combs in enumerate(combinations):
        X_new[:, i] = np.prod(X[:, index_combs], axis=-1)

    return X_new


def normalize(X):
    X[:, 1:] = (X[:, 1:] - np.mean(X[:, 1:], axis=0)) / np.std(
        X[:, 1:], axis=0
    )
    return X



import logging

logger = logging.getLogger(__name__)
params = dict()

try:
    import numpy
except ImportError:
    raise ImportError(
        "'swi-ml has a single NumPy dependency, visit their installation "
        "guide: https://numpy.org/install/"
    )

try:
    import cupy

    _raise_cupy_error = False
except ImportError:
    _raise_cupy_error = True
    logger.warning(
        "No 'cupy' installation found, backend will be defaulted to 'numpy'"
    )


class _Backend:
    def __init__(self):
        global params
        self.backend = None

    def set_backend(self, backend):
        global params
        logger.warning(f"Setting backend: {backend}")
        if backend == "numpy":
            params["backend"] = numpy
        elif backend == "cupy":
            from swi_ml import _fallback_to_numpy

            if not _raise_cupy_error:
                params["backend"] = cupy
            elif _fallback_to_numpy:
                logger.warning(
                    "'cupy' backend not found, falling back to 'numpy'"
                )
                self.set_backend("numpy")
            else:
                raise ImportError(
                    "'cupy' backend needs to be installed first, visit "
                    "https://docs.cupy.dev/en/stable/install.html#install-cupy"
                )

        else:
            raise NotImplementedError(
                "Only 'numpy' and 'cupy' backends are supported"
            )
        self.backend = params["backend"]

    def get_backend(self):
        global params
        if "backend" not in params.keys():
            logger.critical("Backend is not set, using default 'numpy'")
            self.set_backend("numpy")
        return params["backend"]



import logging

from swi_ml import logger as _global_logger
from swi_ml.backend import _Backend
from swi_ml import distributions

logger = logging.getLogger(__name__)


class NaiveBayesClassification(_Backend):
    def __init__(self, distribution="gaussian", verbose=None):
        if verbose is not None:
            logger.setLevel(verbose)
        else:
            logger.setLevel(_global_logger.getEffectiveLevel())

        if distribution.lower() == "gaussian":
            self.distribution = distributions.GaussianDistribution
        else:
            raise NotImplementedError(
                "Only 'gaussian' distribution is implemented yet"
            )
        self.unique_classes = None
        self._class_distributions = dict()
        self.backend = super().get_backend()

    def _calculate_class_prior(self, unique_class):
        """
        prior of unique_class
        (samples_with_class == unique_class / total_samples)
        """
        return self.backend.mean(self.unique_classes == unique_class)

    def _fit_preprocess(self, data, labels):
        # cast to array, CuPy backend will load the arrays on GPU
        X = self.backend.asarray(data)
        Y = self.backend.asarray(labels)
        return X, Y

    def fit(self, data, labels):
        """
        Learns the distribution of each label for each feature
        """
        X, Y = self._fit_preprocess(data, labels)
        self.unique_classes = self.backend.unique(Y).astype(int)
        logger.info(f"Unique Classes: {self.unique_classes.shape}")

        for label in self.unique_classes:
            data_subset = X[self.backend.where(Y == label)]
            # to prevent unhashable type (see https://github.com/cupy/cupy/issues/4531)
            label = int(label)
            # bootstrap class distribution as list (of distributions)
            self._class_distributions[label] = []
            for feature in data_subset.T:
                # get distribution from mean and variance
                dist = self.distribution(feature.mean(), feature.var())
                logger.info(f"Label: {label}, Distribution: {dist}")
                self._class_distributions[label].append(dist)

    def _predict_preprocess(self, data, probability):
        data = self.backend.asarray(data)
        if len(data.shape) < 2:
            data = data.reshape(1, -1)
        if probability:
            predictions = self.backend.empty((data.shape[0], 2))
        else:
            predictions = self.backend.empty(data.shape[0])
        return data, predictions

    def predict(self, data, probability=True):
        """
        Classifies samples according to the largest probability
        with the learnt distributions
        """
        data, predictions = self._predict_preprocess(data, probability)
        for idx, sample in enumerate(data):
            class_probabilities = self.backend.empty(self.unique_classes.shape)
            for unique_class in self.unique_classes:
                posteriors = self._calculate_class_prior(unique_class)
                # to prevent unhashable type (see https://github.com/cupy/cupy/issues/4531)
                unique_class = int(unique_class)
                class_distributions = self._class_distributions[unique_class]
                for sample_value, dist in zip(sample, class_distributions):
                    posteriors *= dist.pdf(sample_value)
                class_probabilities[unique_class] = posteriors
            if probability:
                predictions[idx] = class_probabilities
            else:
                predictions[idx] = self.unique_classes[
                    self.backend.argmax(class_probabilities)
                ]
        return predictions



import pytest
from numpy.testing import assert_equal
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split

from swi_ml import (
    logger,
    set_automatic_fallback,
    set_backend,
    set_logging_level,
)
from swi_ml.classification import (
    NaiveBayesClassification,
)

set_logging_level("DEBUG")
logger.setLevel("DEBUG")

X, Y = make_classification(n_samples=1000, n_features=5, random_state=42)

X_train, X_test, Y_train, Y_test = train_test_split(
    X, Y, test_size=1 / 3, random_state=0
)


def test_naive_bayes():
    dist = "gaussian"

    # set numpy backend first
    set_backend("numpy")
    model_np = NaiveBayesClassification(distribution=dist)

    model_np.fit(X_train, Y_train)
    Y_pred_np = model_np.predict(X_test, probability=False)

    with pytest.raises(NotImplementedError):
        _ = NaiveBayesClassification(
            distribution="wrong_distribution",
        )

    assert Y_pred_np is not None


def test_naive_bayes_probability():
    dist = "gaussian"

    model_np = NaiveBayesClassification(distribution=dist)

    model_np.fit(X_train, Y_train)
    Y_pred = model_np.predict(X_test, probability=False)
    Y_pred_proba = model_np.predict(X_test, probability=True)

    assert Y_pred.shape[0] == Y_pred_proba.shape[0]
    assert Y_pred_proba.shape[1] == 2


@pytest.mark.xfail
def test_naive_bayes_special_case():
    dist = "gaussian"

    # set cupy backend
    set_backend("cupy")

    # special case: f"{element}:.2f" is unsupported for CuPy backend
    # fixed in https://github.com/cupy/cupy/issues/4532
    with pytest.raises(TypeError):
        model_cp = NaiveBayesClassification(distribution=dist)

        model_cp.fit(X_train, Y_train)

        Y_pred_cp = model_cp.predict(X_test)

        set_backend("numpy")
        model_np = NaiveBayesClassification(distribution=dist)

        model_np.fit(X_train, Y_train)
        Y_pred_np = model_np.predict(X_test, probability=False)

        assert Y_pred_np.shape == Y_pred_cp.shape



from swi_ml.backend import _Backend


class Sigmoid(_Backend):
    def __init__(self):
        self.backend = super().get_backend()

    def activate(self, X):
        return 1 / (1 + self.backend.exp(-X))



import math
from swi_ml.backend import _Backend


class GaussianDistribution(_Backend):
    """
    Gaussian Distribution Class
    """

    def __init__(self, mean, variance):
        self.mean = mean
        self.var = variance
        self.backend = super().get_backend()

    def pdf(self, x):
        """
        Returns likelihood of sample according to the distribution
        """
        # prevent division by zero
        eps = 1e-4
        coefficient = 1 / math.sqrt(2 * math.pi * self.var + eps)
        exponent = math.exp(
            -(math.pow(x - self.mean, 2) / (2 * self.var + eps))
        )
        return coefficient * exponent

    def __repr__(self):
        """
        Represents the characteristics of the distribution
        """
        return f"Gaussian [Mean: {self.mean:.2f}, Variance: {self.var:.2f}]"



import logging
import math
import time

from swi_ml.backend import _Backend
from swi_ml import logger as _global_logger, manipulations

logger = logging.getLogger(__name__)


class SVM(_Backend):
    def __init__(
        self,
        num_iterations: int,
        learning_rate: float,
        regularisation_ratio=0.5,
        hinge_constant=1.0,
        normalize=False,
        regularisation=None,
        initialiser="uniform",
        verbose=None,
    ):
        if verbose is not None:
            logger.setLevel(verbose)
        else:
            logger.setLevel(_global_logger.getEffectiveLevel())
        self.num_iterations = num_iterations
        self.learning_rate = learning_rate
        self.regularisation_ratio = regularisation_ratio
        self.hinge_constant = hinge_constant
        self.normalize = normalize
        self.initialiser = initialiser
        self.history = []
        self.backend = super().get_backend()

    def _initialise_uniform_weights(self, shape: tuple):
        self.num_samples, self.num_features = shape
        limit = 1 / math.sqrt(self.num_features)
        self.W = self.backend.asarray(
            self.backend.random.uniform(-limit, limit, (self.num_features,))
        )
        self.b = self.backend.zeros(
            1,
        )
        self.hinge_constant = self.backend.asarray(
            self.hinge_constant,
        )

    def _initialise_zeros_weights(self, shape: tuple):
        self.num_samples, self.num_features = shape
        self.W = self.backend.asarray(
            self.backend.zeros(
                self.num_features,
            )
        )
        self.b = self.backend.zeros(
            1,
        )

    def initialise_weights(self, X):
        """
        Initialises weights with correct dimensions
        """
        if self.initialiser == "uniform":
            self._initialise_uniform_weights(X.shape)
        elif self.initialiser == "zeros":
            self._initialise_zeros_weights(X.shape)
        else:
            raise NotImplementedError(
                "Only 'uniform' and 'zeros' initialisers are supported"
            )

    def hinge_loss(self, Y, Y_pred):
        mul = self.backend.multiply(Y, Y_pred)
        sub = self.backend.subtract(self.hinge_constant, mul)
        return self.backend.mean(self.backend.where(sub < 0, 0, sub))

    def _fit_preprocess(self, data, labels):
        if (
            self.regularisation_ratio <= 0.0
            or self.regularisation_ratio >= 1.0
        ):
            raise ValueError("regularisation ratio should be between 0 and 1")

        # offload to-be-used-once tasks to CPU
        if self.normalize:
            data = manipulations.normalize(data)
        # cast to array, CuPy backend will load the arrays on GPU
        X = self.backend.asarray(data)
        Y = self.backend.asarray(labels)
        return X, Y

    def fit(self, data, labels):
        """
        Given data and labels, it runs the actual training logic
        """
        logger.debug(f"Current Backend: {self.backend}")
        X, Y = self._fit_preprocess(data, labels)

        self.initialise_weights(X)

        start = time.time()
        for it in range(self.num_iterations):
            Y_pred = self._predict(X, hinge=True)

            self.curr_loss = self.hinge_loss(Y, Y_pred)
            logger.info(
                f"Hinge Loss ({it+1}/{self.num_iterations}): {self.curr_loss}"
            )

            safe_zones = self.backend.where(
                Y <= 0, -self.hinge_constant, self.hinge_constant
            )

            for idx, sample in enumerate(X):
                if (
                    safe_zones[idx] * self._predict(sample)
                    >= self.hinge_constant
                ):
                    self._dW = 2 * self.regularisation_ratio * self.W
                else:
                    self._dW = 2 * self.regularisation_ratio * self.W - (
                        1 - self.regularisation_ratio
                    ) * sample.dot(safe_zones[idx])
                self._db = safe_zones[idx]

                self._update_weights()
            self._update_history()

        logger.info(f"Training time: {time.time()-start} seconds")

        return self

    def _update_weights(self):
        self.W = self.W - self.learning_rate * self._dW
        self.b = self.b - self.learning_rate * self._db
        return self

    def _predict(self, X, hinge=False):
        if hinge:
            return X.dot(self.W) - self.b
        else:
            return self.backend.sign(X.dot(self.W) - self.b)

    def _predict_preprocess(self, data):
        if self.normalize:
            data = manipulations.normalize(data)
        return self.backend.asarray(data)

    def predict(self, X, hinge=False):
        """
        Given an input array X, it returns the prediction array
        (GPU array if CuPy backend is enabled) after inferencing
        """
        return self._predict(self._predict_preprocess(X), hinge=hinge)

    def _update_history(self):
        self.history.append(self.curr_loss)



import pytest
from numpy.testing import assert_equal
from sklearn.datasets import make_regression
from sklearn.model_selection import train_test_split

from swi_ml import (
    logger,
    set_automatic_fallback,
    set_backend,
    set_logging_level,
)
from swi_ml.regression import (
    LinearRegressionGD,
    LassoRegressionGD,
    RidgeRegressionGD,
    ElasticNetRegressionGD,
    PolynomialRegressionGD,
)

set_logging_level("DEBUG")
logger.setLevel("DEBUG")

X, Y = make_regression(n_samples=1000, n_features=5, random_state=42)

X_train, X_test, Y_train, Y_test = train_test_split(
    X, Y, test_size=1 / 3, random_state=0
)


def test_linear_regression():
    # only test which tests both NumPy and CuPy backends
    num_iters = 10
    lr = 0.05

    # set numpy backend first
    set_backend("numpy")
    model_np = LinearRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
    )

    with pytest.raises(ValueError):
        model_np.fit([], [])

    model_np.fit(X_train, Y_train)
    Y_pred_np = model_np.predict(X_test)

    with pytest.raises(NotImplementedError):
        wrong_initialiser_model = LinearRegressionGD(
            num_iterations=num_iters,
            learning_rate=lr,
            initialiser="wrong_initialiser",
        )
        wrong_initialiser_model.fit(X_train, Y_train)

    # set cupy backend
    set_backend("cupy")
    model_cp = LinearRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
    )

    model_cp.fit(X_train, Y_train)

    Y_pred_cp = model_cp.predict(X_test)

    assert len(model_np.history) == len(model_cp.history) == num_iters
    assert Y_pred_np.shape == Y_pred_cp.shape

    # switch backend back to numpy
    set_backend("numpy")


def test_lasso_regression():
    num_iters = 10
    lr = 0.05

    # needs l1_cost
    with pytest.raises(TypeError):
        _ = LassoRegressionGD(
            num_iterations=num_iters,
            learning_rate=lr,
        )

    model = LassoRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
        l1_cost=100,
    )

    model.fit(X_train, Y_train)

    _ = model.predict(X_test)

    assert model.MSE_loss is not None
    assert model.regularisation is not None


def test_ridge_regression():
    num_iters = 10
    lr = 0.05

    # needs l2_cost
    with pytest.raises(TypeError):
        _ = RidgeRegressionGD(
            num_iterations=num_iters,
            learning_rate=lr,
        )

    model = RidgeRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
        l2_cost=100,
    )

    model.fit(X_train, Y_train)

    _ = model.predict(X_test)

    assert model.MSE_loss is not None
    assert model.regularisation is not None


def test_elasticnet_regression():
    num_iters = 10
    lr = 0.05

    # needs multiplying factor, as well as l1_ratio
    with pytest.raises(TypeError):
        _ = ElasticNetRegressionGD(
            num_iterations=num_iters,
            learning_rate=lr,
        )

    # test l1_ratio, initialiser set to zero to have same initial weights
    elastic_net_lasso_model = ElasticNetRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
        multiply_factor=100,
        l1_ratio=1,
        initialiser="zeros",
        normalize=False,
    )
    elastic_net_ridge_model = ElasticNetRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
        multiply_factor=100,
        l1_ratio=0,
        initialiser="zeros",
        normalize=False,
    )

    lasso_model = LassoRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
        l1_cost=100,
        initialiser="zeros",
        normalize=False,
    )
    ridge_model = RidgeRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
        l2_cost=100,
        initialiser="zeros",
        normalize=False,
    )

    elastic_net_lasso_model.fit(X_train, Y_train)
    elastic_net_lasso_pred = elastic_net_lasso_model.predict(X_test)

    elastic_net_ridge_model.fit(X_train, Y_train)
    elastic_net_ridge_pred = elastic_net_ridge_model.predict(X_test)

    lasso_model.fit(X_train, Y_train)
    lasso_pred = lasso_model.predict(X_test)

    ridge_model.fit(X_train, Y_train)
    ridge_pred = ridge_model.predict(X_test)

    assert_equal(elastic_net_lasso_pred, lasso_pred)
    assert_equal(elastic_net_ridge_pred, ridge_pred)


def test_polynomial_regression():
    num_iters = 10
    lr = 0.05

    # needs degree
    with pytest.raises(TypeError):
        _ = PolynomialRegressionGD(
            num_iterations=num_iters,
            learning_rate=lr,
        )

    # test degree
    linear_regression_model = LinearRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
        initialiser="zeros",
        normalize=False,
    )
    polynomial_regression_model = PolynomialRegressionGD(
        num_iterations=num_iters,
        learning_rate=lr,
        initialiser="zeros",
        normalize=False,
        degree=1,
    )

    linear_regression_model.fit(X_train, Y_train)
    linear_regression_pred = linear_regression_model.predict(X_test)

    polynomial_regression_model.fit(X_train, Y_train)
    polynomial_regression_pred = polynomial_regression_model.predict(X_test)

    # assert_equal(polynomial_regression_pred, linear_regression_pred)

    assert polynomial_regression_model.MSE_loss is not None
    assert polynomial_regression_model.degree is not None



from swi_ml.backend import _Backend


class _BaseRegularisation(_Backend):
    """
    Base Class for Regularisation, L1 and L2 regularisations inherit from this
    NOTE: Can be used directly as a L1_L2 (ElasticNet) Regularisation
    """

    def __init__(self, multiply_factor: float, l1_ratio: float):
        self.multiply_factor = (
            multiply_factor if multiply_factor is not None else 1
        )
        self.l1_ratio = l1_ratio if l1_ratio is not None else 1
        self.backend = super().get_backend()

    def add_cost_regularisation(self, W):
        l1_regularisation = self.l1_ratio * self.backend.linalg.norm(W)
        l2_regularisation = (1 - self.l1_ratio) * W.T.dot(W)
        return self.multiply_factor * (l1_regularisation + l2_regularisation)

    def add_gradient_regularisation(self, W):
        l1_regularisation = self.l1_ratio * self.backend.sign(W)
        l2_regularisation = (1 - self.l1_ratio) * 2 * W
        return self.multiply_factor * (l1_regularisation + l2_regularisation)


class L1Regularisation(_BaseRegularisation):
    """
    Lasso Regression Regularisation
    """

    def __init__(self, l1_cost: float):
        multiply_factor = l1_cost
        l1_ratio = 1
        super().__init__(multiply_factor, l1_ratio)


class L2Regularisation(_BaseRegularisation):
    """
    Ridge Regression Regularisation
    """

    def __init__(self, l2_cost: float):
        multiply_factor = l2_cost
        l1_ratio = 0
        super().__init__(multiply_factor, l1_ratio)


class L1_L2Regularisation(_BaseRegularisation):
    """
    ElasticNet Regression Regularisation
    """

    def __init__(self, multiply_factor: float, l1_ratio: float):
        super().__init__(multiply_factor, l1_ratio)



import logging
import math
import time

import matplotlib.pyplot as plt

from swi_ml import logger as _global_logger
from swi_ml.backend import _Backend
from swi_ml.manipulations import normalize, transform_polynomial
from swi_ml.regularisers import (
    _BaseRegularisation,
    L1Regularisation,
    L2Regularisation,
    L1_L2Regularisation,
)

logger = logging.getLogger(__name__)


class _BaseRegression(_Backend):
    """
    Base Class for Regression, all regression classes inherit from this
    """

    def __init__(
        self,
        num_iterations: int,
        learning_rate: float,
        normalize=False,
        regularisation=None,
        initialiser="uniform",
        verbose=None,
    ):
        if verbose is not None:
            logger.setLevel(verbose)
        else:
            logger.setLevel(_global_logger.getEffectiveLevel())
        self.num_iterations = num_iterations
        self.learning_rate = learning_rate
        self.normalize = normalize
        self.initialiser = initialiser
        self.regularisation = regularisation
        self.history = []
        self.backend = super().get_backend()

    def _initialise_uniform_weights(self, shape: tuple):
        self.num_samples, self.num_features = shape
        limit = 1 / math.sqrt(self.num_features)
        self.W = self.backend.asarray(
            self.backend.random.uniform(-limit, limit, (self.num_features,))
        )
        self.b = self.backend.zeros(
            1,
        )

    def _initialise_zeros_weights(self, shape: tuple):
        self.num_samples, self.num_features = shape
        self.W = self.backend.asarray(
            self.backend.zeros(
                self.num_features,
            )
        )
        self.b = self.backend.zeros(
            1,
        )

    def _update_history(self):
        self.history.append(self.curr_loss)

    def _update_weights(self):
        self.W = self.W - self.learning_rate * self._dW
        self.b = self.b - self.learning_rate * self._db
        return self

    def MSE_loss(self, Y_true, Y_pred):
        """
        Given ground truth and predicted values, it returns the Mean Square Error
        """
        return self.backend.mean(
            0.5 * (Y_true - Y_pred) ** 2
        ) + self.regularisation.add_cost_regularisation(self.W)

    def initialise_weights(self, X):
        """
        Initialises weights with correct dimensions
        """
        if self.initialiser == "uniform":
            self._initialise_uniform_weights(X.shape)
        elif self.initialiser == "zeros":
            self._initialise_zeros_weights(X.shape)
        else:
            raise NotImplementedError(
                "Only 'uniform' and 'zeros' initialisers are supported"
            )

    def _fit_preprocess(self, data, labels):
        # offload to-be-used-once tasks to CPU
        if self.normalize:
            data = normalize(data)
        # cast to array, CuPy backend will load the arrays on GPU
        X = self.backend.asarray(data)
        Y = self.backend.asarray(labels)
        return X, Y

    def fit(self, data, labels):
        """
        Given data and labels, it runs the actual training logic
        """
        logger.debug(f"Current Backend: {self.backend}")
        X, Y = self._fit_preprocess(data, labels)

        self.initialise_weights(X)

        start = time.time()
        for it in range(self.num_iterations):
            Y_pred = self._predict(X)
            diff = self.backend.subtract(Y, Y_pred)

            self.curr_loss = self.MSE_loss(Y, Y_pred)
            logger.info(
                f"MSE ({it+1}/{self.num_iterations}): {self.curr_loss}"
            )
            # regularisation magnitude is 0 for vanilla linear regression
            self._dW = (
                -(
                    2 * (X.T).dot(diff)
                    - self.regularisation.add_gradient_regularisation(self.W)
                )
                / self.num_samples
            )
            self._db = -2 * self.backend.sum(diff) / self.num_samples

            self._update_weights()
            self._update_history()

        logger.info(f"Training time: {time.time()-start} seconds")

        return self

    def _predict_preprocess(self, data):
        if self.normalize:
            data = normalize(data)
        return self.backend.asarray(data)

    def predict(self, X):
        """
        Given an input array X, it returns the prediction array
        (GPU array if CuPy backend is enabled) after inferencing
        """
        return self._predict(self._predict_preprocess(X))

    def _predict(self, X):
        return X.dot(self.W) + self.b

    def plot_loss(self):
        """
        Plots the loss history curve during the training period.
        NOTE: This function just plots the graph, to display it
        explicitly call `matplotlib.pyplot.show()`
        """
        plt.title("MSE History")
        plt.xlabel("num_iterations")
        plt.ylabel("MSE")
        plt.legend()


class LinearRegressionGD(_BaseRegression):
    def __init__(
        self,
        num_iterations: int,
        learning_rate: float,
        normalize=False,
        initialiser="uniform",
        verbose=None,
    ):
        # regularisation of alpha 0 (essentially NIL)
        regularisation = _BaseRegularisation(multiply_factor=0, l1_ratio=0)
        super().__init__(
            num_iterations,
            learning_rate,
            normalize,
            regularisation,
            initialiser,
            verbose,
        )

    def plot_loss(self):
        plt.plot(self.history, label="Linear Regression")
        super().plot_loss()


class LassoRegressionGD(_BaseRegression):
    def __init__(
        self,
        num_iterations: int,
        learning_rate: float,
        l1_cost: float,
        normalize=False,
        initialiser="uniform",
        verbose=None,
    ):
        regularisation = L1Regularisation(l1_cost=l1_cost)
        super().__init__(
            num_iterations,
            learning_rate,
            normalize,
            regularisation,
            initialiser,
            verbose,
        )

    def plot_loss(self):
        plt.plot(self.history, label="Lasso Regression")
        super().plot_loss()


class RidgeRegressionGD(_BaseRegression):
    def __init__(
        self,
        num_iterations: int,
        learning_rate: float,
        l2_cost: float,
        normalize=False,
        initialiser="uniform",
        backend="cupy",
        verbose=None,
    ):
        regularisation = L2Regularisation(l2_cost=l2_cost)
        super().__init__(
            num_iterations,
            learning_rate,
            normalize,
            regularisation,
            initialiser,
            verbose,
        )

    def plot_loss(self):
        plt.plot(self.history, label="Ridge Regression")
        super().plot_loss()


class ElasticNetRegressionGD(_BaseRegression):
    def __init__(
        self,
        num_iterations: int,
        learning_rate: float,
        multiply_factor: float,
        l1_ratio: float,
        normalize=False,
        initialiser="uniform",
        verbose=None,
    ):
        regularisation = L1_L2Regularisation(
            multiply_factor=multiply_factor, l1_ratio=l1_ratio
        )
        super().__init__(
            num_iterations,
            learning_rate,
            normalize,
            regularisation,
            initialiser,
            verbose,
        )

    def plot_loss(self):
        plt.plot(self.history, label="Elastic Net Regression")
        super().plot_loss()


class PolynomialRegressionGD(_BaseRegression):
    def __init__(
        self,
        num_iterations: int,
        learning_rate: float,
        degree: float,
        multiply_factor=None,
        l1_ratio=None,
        normalize=False,
        initialiser="uniform",
        verbose=None,
    ):
        self.degree = degree
        regularisation = L1_L2Regularisation(
            multiply_factor=multiply_factor, l1_ratio=l1_ratio
        )
        super().__init__(
            num_iterations,
            learning_rate,
            normalize,
            regularisation,
            initialiser,
            verbose,
        )

    def _fit_preprocess(self, data, labels):
        poly_data = transform_polynomial(data, self.degree)
        return super()._fit_preprocess(poly_data, labels)

    def _predict_preprocess(self, data):
        poly_data = transform_polynomial(data, self.degree)
        return super()._predict_preprocess(poly_data)

    def plot_loss(self):
        plt.plot(
            self.history, label=f"Polynomial Regression, degree={self.degree}"
        )
        super().plot_loss()
